/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#ifndef AND_SCR_H
#define AND_SCR_H

typedef void (*write_proc_t)(unsigned char *,int,int,int,int);

int SCR_UpCaseRepresentation(int key_code);

void SCR_CursorOn (void);

int SCR_CursorOff (void);

void SCR_SetCursorPosn(int row, int col);

void SCR_Write(unsigned char *data,int row, int col, int len, int attr);

void SCR_Box(int rows, int cols, int ypos, int xpos, int attr, write_proc_t write_proc);

void SCR_Line(int rows, int cols, int ypos, int xpos, int attr, write_proc_t write_proc);

void SCR_RestoreVideoEnv (void);

void SCR_ClearStatusLine(void);

void SCR_WriteStatusLine(char *str);

int SCR_UpCaseRepresentation(int key_code);

void SCR_Beep (void);

int SCR_Keyboard_io(void);

int SCR_ClearKeyboard(void);

int SCR_GetKey(void);

int CheckKeypad(int chin, int prefixed);

void SuspendProcess(void);

void SCR_SetupVideo(void);

char *SCR_GetCap(char *a_cap, char *buf);

void SCR_UseCharSet(int num);

void SCR_InitCharMap(char hl, char vl, char ul, char ll, char ur, 
		    char lr, char lb, char rb);

void SCR_NoRawState(void);

void SCR_RawState(void);

void SCR_SetStringAttribute(char *attr_string,
			    int nlines);

void SCR_RedrawScreen(void);

void SCR_UpdateScreen(void);

int SCR_AlternateCharSet(int on_off);

void SCR_SetBlink(int on_off);

int SCR_CanHighlight(void);

void SCR_SetHighlight(int on_off);

void SCR_ClearScreen(void);

void SCR_LeftT(int r,int c,int attr,write_proc_t write_proc);

void SCR_RightT(int r,int c,int attr,write_proc_t write_proc);

void SCR_NoHighlight(void);

void SCR_SendIS(void);

int get_itimer_value(void);

#endif /* AND_SCR_H */
