/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef MENUS_H
#define MENUS_H

/*
 * menu definitions for the mail client
 */

#include "common.h"

struct menu_loc_struct {
    unsigned char submenu_size;
    unsigned char pos;
    unsigned char size;
    unsigned char box_width;
    unsigned char box_height;
    unsigned char pop_col;
    unsigned char pop_row;
};
typedef struct menu_loc_struct MENULOC, *MENULOC_pt;

struct menu_struct {
    int	    id;
    char    *name;
    Boolean enabled;
    struct  menu_struct *submenu;
    MENULOC *loc;
    Boolean initial;
    int     bykey;
    Boolean check;
};
typedef struct menu_struct MENU, *MENU_pt;

struct menu_tree_struct {
  int      id;
  MENU    *menu;
  char    *name;
  Boolean  enabled;
  MENULOC *loc;
};
typedef struct menu_tree_struct	MENU_TREE, *MENU_TREE_pt;

/* menu items greater than 2000 are assumed to be remote */

/* definitions for main menus */

#define M_EZMAIL    1	/* for local, use 100-199 */
#define M_COMPOSE   2	/* for local, use 200-299 */
#define M_EDIT      3	/* for local, use 600-699 */
#define M_FOLDERS   4	/* for local, use 300-399 */
#define M_MESSAGE   5	/* for local, use 400-499 */
#define M_SERVICES  6	/* for local, use 500-599 */
#define M_WINDOW    7	/* for local, use 700-799 */

/* Submenus under EZMAIL */

#define M_ABOUT		100
#define M_HELP		101
#define M_BUG		102
#define M_KEYMAP        103
#define M_QUIT		104

/* Submenus under WINDOW */

#define M_BRING_FRONT      700
#define    M_SHOW_FOLDERS  701
#define    M_SHOW_CAPTIONS 702
#define    M_SHOW_MESSAGE  703
#define    M_SHOW_CONSOLE  704
#define    M_SHOW_DOCUMENT 705
#define    M_SHOW_FILER    706
#define M_MOVE             707
#define M_RESIZE           708
#define M_ZOOM             709
#define M_CLOSE            710
#define M_SAVE             711
#define	    M_SAVE_LOCAL   712
#define	    M_SAVE_REMOTE  713
#define M_PRINT_LOCAL      714
#define M_REDRAW           715

/* submenus under COMPOSE */

#define M_NEW	        	200
#define M_OPEN			201
#define	    M_OPEN_LOCAL	202
#define	    M_OPEN_REMOTE	2107
#define M_APPEND		203
#define	    M_APND_LOCAL	204
#define	    M_APND_REMOTE	2104
#define M_MAIL_HEADERS          213
#define     M_ADD_FAXHDRS       214
#define     M_ADD_HEADERS       215
#define     M_ADD_ACK           216
#define     M_ADD_VOTE          217
#define     M_ADD_SUB_INVITE    218
#define     M_ADD_DIST_INVITE   219
#define     M_ADD_REPLY_TO      220
#define M_SEND			2100
#define M_VERIFY		2101
#define M_WHOIS		        2102
#define M_SIGNMSG		2103

/* submenus under EDIT    */

#define M_EDITMSG               600
#define M_EXTERNAL_EDIT         601
#define M_CUT                   602
#define M_COPY                  603
#define M_PASTE                 604
#define M_PREFERENCES           606
#define    M_LOCAL_PREFS        607
#define    M_REMOTE_PREFS       608
#define    M_SIGN_THIS_MAIL     609
#define    M_BCC_THIS_MAIL      610
#define M_SET_MARK              611

/* submenus under FOLDERS */

#define M_CREATE		2000
#define M_DELETE		2001
#define M_RENAME		2002
#define M_MERGE			2003
#define M_INFO			2004
#define M_EXPOSE		300
#define	    M_EXPOSE_MATCH	2010
#define	    M_EXPOSE_PERSONAL	2011
#define	    M_EXPOSE_ALL	2012
#define	    M_EXPOSE_SUBSCRIBED	2013
#define	    M_EXPOSE_CHANGED	2014
#define     M_EXPOSE_BYNAME     2019
#define M_SUBSCRIBE		301
#define	    M_SUB_FULL		2006
#define	    M_SUB_NONE		2007
#define	    M_SUB_SHOWALL	2008
#define	    M_SUB_PRINT		2015
#define	    M_SUB_ASK		2009
#define M_SEEN			2005
#define M_POST			2016
#define M_PUNT			2018
#define M_QUOTA			302

/* submenus under MESSAGE */

#define M_KILL			2200
#define M_NEXT			2201
#define M_PREV			2202
#define M_MDELETE		2203
#define M_UNDELETE		2204
#define M_RESPOND		400
#define	    M_REPLY_SENDER	2209
#define	    M_REPLY_READERS	2210
#define	    M_REPLY_RESEND	2211
#define	    M_REPLY_FORWARD	2212
#define	    M_REPLY_ALL		2213
#define     M_REPLY_DRAFT       2220
#define M_C_POST                2214
#define M_FILEINTO		401
#define	    M_FILE_BYNAME	2205
#define M_RESTORE		2207
#define M_SETNEW		2208
#define M_HEADERS		2206
#define M_PRINT_REMOTE          2215
#define M_FILE_IN_FILE          2216
#define M_ROT_13                2217
#define M_EXPUNGE               2218
#define M_EXCERPT               402
#define M_SEARCH                2219

/* submenus under services */

#define M_GETMAIL		2017
#define M_CONNECT		500
#define M_DISCONNECT		501
#define M_OTHERSVC		502
#define     M_FINGER            2112
#define	    M_SVC_PASSWORD	2111
#define	    M_SVC_FORWARD	2110
#define     M_MOTD		2106
#define     M_AFS_PRINTQ        2116
#define     M_AFS_HELP          2117
#define M_AFS_SERVICES          505
#define     M_AFS_RM            2300
#define     M_AFS_MV            2301
#define     M_AFS_EDIT          2302
#define     M_AFS_CAT           2303
#define     M_AFS_HOME          2306
#define     M_AFS_MKDIR         2308
#define     M_AFS_UPDATE        2309
#define     M_AFS_LONGLIST      2311
#define     M_AFS_CD            2312
#define     M_AFS_QUOTA         2314
#define     M_AFS_FILEINFO      2315
#define     M_AFS_NEW           2316
#define M_QUICK_EDIT            2313
#define     M_QE_ALIASES        2710
#define     M_QE_PREFS          508
#define     M_QE_DOTPREFS       509
#define     M_QE_SIGFILE        2713
#define M_UPDATE	        2109

#define REMOTE_FILEINTO -5000

/* Add seperator lines to menus */

#define WANT_NULL_MENU

#ifdef WANT_NULL_MENU
#define NULL_MENU 0,"",FALSE,(MENU *)NULL,(MENULOC *)NULL,0,0,FALSE,
#else
#define NULL_MENU
#endif

void MNU_CheckItem(int item, int on_off);

void MNU_EnableBar(int bar_item, int aval);

void MNU_EnableItem(int item, int aval);

void MNU_RevertToInitial(void);

void MNU_EnableAll(int aval);

void MNU_GetMenuName(int mid,char *astr);

void MNU_initialize(int (*do_menu_command_proc)());

void MNU_SetupMenuBar(void);

void MNU_SetKeyCode(int item, int keycode);

void MNU_SetMenuCmdProc(int (*aProc)());

void MNU_CloseMenu(int all);

void MNU_DrawMenuBox(MENULOC *amenuloc,Boolean *focus);

void MNU_PullDownMenu(void);

void pull_it_down(MENU *amenu,MENULOC *amenuloc,Boolean *focus);

void MNU_PullDownHMenu(MENU *amenu);

int MNU_FindMenuByKey(int keycode);

void MNU_ReturnControl(int do_cmd);

void MNU_SelectMenuBar(int keycode);

void MNU_MenuMode(int keycode);

void MNU_ResetFileInto(void);

void MNU_AddFileinto(char *name,
		     int add_or_remove);



#endif /* MENUS_H */
