/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <varargs.h>


#include "pfio.h"


/*
 * pack arguments for network transfer
 */
void PFpack(va_alist)
va_dcl
{
    va_list ap;
    int atype;
    long along;
    char *astr;
    PFM_pt o;
    va_start(ap);
    o=va_arg(ap,PFM_pt);
    atype=va_arg(ap,int);  
    along=va_arg(ap,int);
    if(atype!=LS_alocop)
	(*(o->parerr))(o,"PFIO:invalid opcode %d",atype);
    PF_aloc(o,(long)along);
    o->buf+=MIN_PAK;		/*skip the length we will come back to it*/
    
    /* pack up the argument list*/
    while(TRUE) {
        if(PFM_remaining(*o)<4)
	  (*(o->parerr))(o,"PFIO:output reserve too small");
	atype=va_arg(ap,int);
   	switch(atype) {
	    case LS_dropoff:
		goto end_of_arg_list;
	    case LS_int32:
		along=va_arg(ap,long);
		*(o->buf++) = along >> 24;
		*(o->buf++) = along >> 16;
		*(o->buf++) = along >> 8;
		*(o->buf++) = along ;
		break;
	    case LS_int16:
		along=va_arg(ap,long);
   		*(o->buf++) = along >> 8;
		*(o->buf++) = along ;
		break;
	    case LS_int8:
		along=va_arg(ap,long);
		*(o->buf++) = along ;
		break;
	    case LS_cstr:
		along=va_arg(ap,long);	/*string length*/
		if(PFM_remaining(*o)<(along+2))
		  (*(o->parerr))(o,"PFIO:output reserve too small");
   		*(o->buf++) = along >> 8;
		*(o->buf++) = along ;
		astr=va_arg(ap,char *);	/*string to send*/
		bcopy(astr,o->buf,((int)along)); /*put on the string*/
		o->buf+=along;	/*and advace output*/
		*(o->buf++)=0;	/*null terminate it*/
		break;
	    case LS_rstr:		/*rest of packet as string*/
		along=va_arg(ap,long);	/*string length*/
		astr=va_arg(ap,char *);	/*string to send*/
		if(PFM_remaining(*o)<(along+2))
		  (*(o->parerr))(o,"PFIO:output reserve too small");
		bcopy(astr,o->buf,((int)along)); /*put on the string*/
		o->buf+=along;	/*and advace output*/
		*(o->buf++)=0;	/*null terminate it*/
		goto end_of_arg_list;
	    default:
		(*(o->parerr))(o,"invalid send opcode %d",atype);
	}
    }

/*
 * all the arguments have been packed
 * backpatch the length
 */
end_of_arg_list:
    PF_dropoff(o);	/*queue this block up for writing*/
}
