/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * rlci urgent and file primatives
 */

#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "rlci.h"
#include "document.h"

void RL_GetBooleanFromUserReply(result, lii)
int result, lii;
{
  RL_header_pt li;
  IF_NET_DEAD{return;}
  li= RLI_to_LI(lii);
		
  /* result == 1 is OK */
  /* result != 1 is CANCEL */
	
  /*  PFpack(
	 &ppfm,
	 LS_aloc(20),
	 LS_seqophead(li->lii,li->serial,SYSM_MASK+SYSM_urgent_reply),
	 LS_msend_num(LS_int8, URGENT_GetBooleanFromUser),
	 LS_msend_num(LS_int8, result),
	 LS_dropoff); */
		
  END_CATCH
}

void RL_ChooseFromListReply(result, lii)
int result, lii;
{

  RL_header_pt li;
  IF_NET_DEAD{return;}	/*catch errors*/
  li = RLI_to_LI(lii);
  
  /*  PFpack(
	 &ppfm,
	 LS_aloc(20),
	 LS_seqophead(li->lii,li->serial,SYSM_MASK+SYSM_urgent_reply),
	 LS_msend_num(LS_int8, URGENT_ChooseFromList),
	 LS_msend_num(LS_int8, result),
	 LS_dropoff); */
	
  END_CATCH
}

void RL_GetStringFromUserReply(theText, lii)
char *theText;
int lii;
{
  int len;
  RL_header_pt li;
  IF_NET_DEAD{return;}	/*catch errors*/
  li = RLI_to_LI(lii);
	
  len = strlen(theText);
  /*  PFpack(
	 &ppfm,
	 LS_aloc(20 + len+1),
	 LS_seqophead(li->lii,li->serial,SYSM_MASK+SYSM_urgent_reply),
	 LS_msend_num(LS_int8, URGENT_GetStringFromUser),
	 LS_msend_rstr(len, theText),
	 LS_dropoff); */
		
  END_CATCH
}

struct fileop_R {
	long	inProgress;     /* a file operation in progress */
	long	size;           /* size of the file */
	long	offset;         /* current offset in file */
	char    *room;          /* address of buffer to read/write */
	long	roomsize;       /* the size of buffer */
	short	last_fop;       /* last file operation requested */
	short	last_error;     /* error from last file operation */
	int	filename_len;   /* filename length of the remote file */
	char	name[1024];     /* name of the remote file */
};
typedef struct fileop_R FILEOP, *FILEOP_pt;

FILEOP	aFile; /* currently we process only one file at a time */

void sysm_file_ops(li,m,mi)
RL_header_pt li;
PFM_pt m;
int mi;
{
  long opcode = 0;
  long len;
  char *where;	
	

	
  switch (opcode) {
  case FOP_size:

    aFile.filename_len = strlen(aFile.name);
    if (aFile.last_fop == FOP_size) aFile.last_fop = 0;
    DOC_PickupFile();
    break;
  case FOP_read:
    if (aFile.last_fop == FOP_read) aFile.last_fop = 0;
    else rl_error("RFile packets are out of sequence");
    where = aFile.room;
    len = PFM_remaining(*m) - 1; /* don't count the null -- LS_rstr */
    if (len > 0) {
      aFile.offset += len;
      while ((len--) > 0) *where++ = *m->buf++;
    }
    DOC_RemoteDataArrived(aFile.room, (int)(where-aFile.room));
    break;
  case FOP_write:
    rl_error("FOP_write received");
    break;
  case FOP_new:

    aFile.filename_len = strlen(aFile.name);
    if ( aFile.last_fop == FOP_new) {
      aFile.last_fop = 0;
      DOC_RemoteFileCreated(aFile.name);
    }
    break;
  case FOP_ack:
    if ( aFile.last_fop == FOP_write) {
      aFile.last_fop = 0;
      DOC_RemoteWriteDone();
    }
    break;
  case FOP_error:
    aFile.last_error = TRUE;
    if (aFile.last_fop) DOC_HandleRFileError();
    break;
  default:
    rl_error("Unknown file operation opcode");
  }
}

int RFile_EOF(void)
{
  if ((aFile.offset >= aFile.size) || (is_net_dead())) return(TRUE);
  return(FALSE);
}

int RFile_error(int clear)
{
  int retval;

  if (is_net_dead() || aFile.last_error) retval = TRUE;
  else retval = FALSE;
  if (clear) aFile.last_error = 0;	
  return(retval);
}

void RFile_set_buf(char *dest,long  size)
{
  aFile.room = dest;
  aFile.roomsize = size;
}

void RFile_get_buf(char **dest)
{
  *dest = &aFile.room[0];
}

long RFile_datasize(void)
{
  return(aFile.size);
}

void RFile_initialize(void)
{
  bzero(&aFile, sizeof(aFile));
	
  aFile.last_fop = FOP_size;	/* in case server asks us to read */
}

char *RFile_filename(void)
{
  return(&aFile.name[0]);
}

void RFile_read(long data_len)
{
  IF_NET_DEAD{return;}	/*catch errors*/
  /*  PFpack(
	 &ppfm,
	 LS_aloc(50 + aFile.filename_len),
	 LS_seqophead(0,0,SYSM_MASK+SYSM_fop),
	 LS_msend_num(LS_int8, FOP_read),
	 LS_msend_num(LS_int32, aFile.offset),
	 LS_msend_num(LS_int32, data_len),
	 LS_msend_str(aFile.filename_len, aFile.name),
	 LS_dropoff); */
  aFile.last_fop = FOP_read;
  END_CATCH
}

void RFile_write(char *data_ptr,long data_len)
{
  IF_NET_DEAD{return;}	/*catch errors*/
  /*  PFpack(
	 &ppfm,
	 LS_aloc(50 + aFile.filename_len + data_len),
	 LS_seqophead(0,0,SYSM_MASK+SYSM_fop),
	 LS_msend_num(LS_int8, FOP_write),
	 LS_msend_num(LS_int32, aFile.offset),
	 LS_msend_num(LS_int32, data_len),
	 LS_msend_str(aFile.filename_len, aFile.name),
	 LS_msend_str(data_len, data_ptr),
	 LS_dropoff); */
		
  aFile.offset += data_len;
  aFile.last_fop = FOP_write;
  END_CATCH
}

void RFile_new(char *fname)
{
  char ourcopy[255];
  int  ourlen = 0;
  IF_NET_DEAD{return;}	/*catch errors*/
  if (fname && (ourlen = strlen(fname)) > 0)
    bcopy(fname, ourcopy, ourlen);
  ourcopy[ourlen] = 0;
  /*  PFpack(
	 &ppfm,
	 LS_aloc(20+ourlen),
	 LS_seqophead(0,0,SYSM_MASK+SYSM_fop),
	 LS_msend_num(LS_int8, FOP_new),
	 LS_msend_rstr(ourlen,ourcopy),
	 LS_dropoff); */
  aFile.last_fop = FOP_new;
  END_CATCH
}

void RFile_remove(void)
{
  IF_NET_DEAD{return;}	/*catch errors*/
  /*  PFpack(
	 &ppfm,
	 LS_aloc(20 + aFile.filename_len+1),
	 LS_seqophead(0,0,SYSM_MASK+SYSM_fop),
	 LS_msend_num(LS_int8, FOP_remove),
	 LS_msend_rstr(aFile.filename_len, aFile.name),
	 LS_dropoff); */
  aFile.last_fop = FOP_remove;
  END_CATCH
}
