/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef RLI_H
#define RLI_H

#include "pfio.h"

extern PFM ppfm;
void rl_parerr();

#define RLI_MAJOR_VERSION (4)
#define RLI_MINOR_VERSION (40)

extern char *c_get_mem();

/* console/system info*/
#define CONSOLE_RL 0
/* user visable folder info */
#define FOLDER_RL 1
/* captions list */
#define CAPTION_RL 2
/* message body */
#define BODY_RL 3
/* send list */
#define SEND_RL 4
/* help index */
#define HELP_RL 5
/* help body */
#define HELPBODY_RL 6
/* directory RL */
#define DIRECTORY_RL 7

#define RL_NUM_LISTS 8

#define RL_LENGTH (86)	/*size of a list entry*/
#define RL_LLENGTH (81) /* logical size of a list entry*/
typedef char RL_entry[RL_LENGTH];

#define RL_TO 0
#define RL_CC 1
#define RL_SUBJECT 2
#define RL_FROM 3
#define RL_DATE 4

struct RL_chunk_R {
  RL_entry cdata;		/*chunk data*/
};
typedef struct RL_chunk_R RL_chunk,*RL_chunk_pt,**RL_chunk_pt_pt;

/* a remote list driver routine */
typedef void (*RL_rtn_pt)();
typedef RL_rtn_pt *RL_rtn_pt_pt;

/*
 * internal macro to call a client or system routine
 */
#define any_CALL(xx_lii,xx_index,xx_kind) \
    (*((RL_driver_tab[xx_lii]+xx_kind)->rtns+xx_index))

/*
 * call the system routine for a given list
 */
#define SYSM_CALL(xx_lii,xx_index) any_CALL(xx_lii,xx_index,MKIND_sys)

/*
 * call the client routine for a given list
 */
#define CLIENT_CALL(xx_lii,xx_index) any_CALL(xx_lii,xx_index,MKIND_clie)

/*
 * return the size of the given client lists message table
 */
#define CLIENT_SIZE(xx_lii) \
  ((RL_driver_tab[xx_lii]+MKIND_clie)->size)

/*
 * list header
 */
struct RL_header_R {
    int	lii;		    /*list number of this list*/
    int	flags;		    /*per list flags*/
    long serial;	    /*list serial number*/
    long size;		    /*number of active entries*/
    long ioloc;		    /*location for next data transmision*/
    long first_pos;	    /*sugjest start position*/
    char *data;		    /*window on the data*/
    long pos;		    /*offset for start of current buffer*/
};
typedef struct RL_header_R RL_header,*RL_header_pt;


RL_header_pt RLI_to_LI();

/*
 * flag bytes in the first byte of each list record
 */
#define RLi_NEW (0x80)		/*client hasn't seen this line*/
#define RLi_VALID (0x40)	/*this line is valid*/

#define LS_ophead(xx_op,xx_lii) \
 (long)(RLC_put_opcode(xx_op)|RLC_put_list(xx_lii)) /*emit the opcode and list*/

#define LS_sendophead(xx_lii,xindex) \
  LS_msend_num(LS_int8,LS_ophead(RLC_opcode_mesg,xx_lii)), \
  LS_msend_num(LS_int16,(xindex))

#define LS_seqophead(xx_lii,xx_seq,xindex) \
  LS_sendophead(xx_lii,xindex), \
  LS_msend_num(LS_int8,(xx_seq))

/*
 * format of transport layer packets (share headers with higher layers)
 * minimum packet length is four bytes
 */
/*
 ********** Begin Format of stream used both for client and server **********
 *
 * The first byte of each message contains an opcode (2 bits) and list number
 * the message is directed to (6 bits).
 */
#define RLC_get_opcode(xx_arg) (((xx_arg)>>5)&0x03)
#define RLC_put_opcode(xx_arg) (((xx_arg)&0x07)<<5)

#define RLC_get_list(xx_arg) ((xx_arg)&0x1f)
#define RLC_put_list(xx_arg) ((xx_arg)&0x1f)

/*
 ********** Begin Format of stream sent to client **********
 */
#define RLC_opcode_mesg 1
#define RLC_opcode_read 2
#define RLC_opcode_read_patch 3

/*
 ********** Begin Format of stream sent to client **********
 */

/* opcodes */
#define RLC_opcode_resize 0	/*resize/set new serial*/
#define RLC_opcode_setloc 1	/*set recieve location*/
#define RLC_opcode_data 2	/*data*/
#define RLC_opcode_prefix 3	/*prefix for seldome used opcodes*/

/* RLC_opcode_data data (bytes
  xyyy yyyy
  if x==0 then data entierly replaces previous contents
          y == first data byte
  if x==1 then data patches previous contents
    int8 == offset for start of patch
  fi
  int8 - length data bytes
*/
#define RLC_data_patch_mask (0x80)

#define NUM_SYSM (8)		/*number of known system messages to lists on server*/
#define SYSM_MASK (0x8000)	/*all system messages have this on*/
#define SYSM_open (0)		/*open a list*/
#define SYSM_hello (1)		/*start a new connection*/
#define SYSM_size (2)		/*set list size*/
#define SYSM_newpc (3)		/*set data transfer offset*/
#define SYSM_urgent (4)		/* urgent reuest, server will block */
#define SYSM_urgent_reply (5)	/* reply to an urgent message */
#define SYSM_fop (6)		/* file operation request */
#define	SYSM_alert (7)		/* alerts */

/* types of urgent messages */

#define URGENT_GetBooleanFromUser	1	/* Get YES or NO from user */
#define URGENT_ChooseFromList		2	/* Choose from a list */
#define URGENT_GetStringFromUser	3	/* Get a string from user */
#define URGENT_Quit			4       /* Signal a shutdown on Mac */
#define URGENT_blurb			5	/* put up a blurb window*/
#define URGENT_set_file_into		6	/*set fileinto menu*/
#define	URGENT_set_prefs		7	/*set preferences*/
#define URGENT_pos_ack			8	/*acknowlege data position*/

#define FOP_size	(1)
#define FOP_read	(2)
#define FOP_write	(3)
#define FOP_new		(4)
#define FOP_remove	(5)
#define FOP_ack		(6)
#define FOP_error	(7)

/* SYSM_alert sub types */

#define	ALERT_warning	(1) /* warning alert */
#define	ALERT_fatal	(2) /* fatal, probably dead */
#define ALERT_console	(3) /*message for console*/
#define ALERT_console_banner (4) /*message for the console banner line*/
#define ALERT_console_beep  (5)  /* Beeps the Mac */
#define ALERT_dialog        (6)  /* Ask for a special dialog */
#define ALERT_hints         (7) /* hints from the server */

#define DIALOG_newmail      (1)   /* You have new mail */
#define DIALOG_idleclose    (2)   /* Closing idle connection */

/*
 * options on RL_new
 */
#define SZ_title 	0x01
#define SZ_banner 	0x02

/*
 * system messages to lists on client
 */
#define RLCS_invalid (0)	/*invalidate a list item*/

#define NUM_MKIND (2)		/*number of kinds of messges (sys,client)*/
#define MKIND_sys (0)
#define MKIND_clie (1)
struct R_rtn_tab_R {
  RL_rtn_pt_pt rtns;		/*table of routines*/
  int size;			/*number of entries*/
};
typedef struct R_rtn_tab_R R_rtn_tab,*R_rtn_tab_pt,**R_rtn_tab_pt_pt;

#define RL_SYSM (3)		/*number remote list system table entries*/

#define RL_sysm_read (0)
#define RL_sysm_open (1)
#define RL_sysm_init (2)

extern void (*rl_incomming_func)();
void RL_read_incomming();

#ifdef IBMPC
#define NUM_OBUFS 4 /* use less output buffers for the dos clients */
#endif

int output_queue_is_full(void);

void set_net_dead(int aval);

void rl_error(char *str);

int is_net_dead(void);

int RL_read(int arli,long idx,char *astr);

void RL_do_num_command(int lii,int cmd,long lindex);
void RL_zero_list(int lii);

void RFile_new(char *fname);
void RFile_remove(void);
void RFile_write(char *data_ptr,long data_len);
void RFile_read(long data_len);
char *RFile_filename(void);
void RFile_initialize(void);
long RFile_datasize(void);
void RFile_get_buf(char **dest);
void RFile_set_buf(char *dest,long  size);
int RFile_error(int clear);
int RFile_EOF(void);
void RL_do_command(int lii,int cmd,long lindex,char *astr);
void set_rl_window_size(int requested_rl_size, int screen_size);


int GetRemoteCommandEntry(int  idx, int listnum,long *strnum, long *instances, 
			  long *needfile,int  *isbusy);
#endif /* RLI_H */
