/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * text buffer handling routines
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>

#include "common.h"
#include "screen.h"
#include "comkeys.h"
#include "menus.h"
#include "text.h"
#include "rli.h"
#include "and_keys.h"
#include "and_scr.h"
#include "edit.h"
#include "and_file.h"
#include "document.h"
#include "dialog.h"
#include "windows.h"
#include "console.h"
#include "and_mach.h"
#include "and_prt.h"
#include "ezconfig.h"

extern char *get_deallocable_mem(), *realloc_deallocable_mem();

extern FILE *FIO_OpenStream();

extern Boolean gCanSendFax;

extern TEXT_VIEW *cons_textview;

int TXT_DoKey();

TEXT_VIEW *text_list = (TEXT_VIEW *)NULL;

#ifdef AFS_ENV
static int Afs_HintKeys[] = {KEYCODE_EDIT, KEYCODE_LOCAL_SAVE};
#else
static int Afs_HintKeys[] = {KEYCODE_EDIT};
#endif

static int Edt_HintKeys[] = {KEYCODE_ESCAPE, KEYCODE_ADD_HEADERS};
static int Snd_HintKeys[] = {KEYCODE_VERIFY, KEYCODE_SEND, KEYCODE_EDIT};
static int Oth_HintKeys[] = {KEYCODE_CLOSE};
static int Gen_HintKeys[] = {KEYCODE_HELP, KEYCODE_KEY_MAP, KEYCODE_QUIT};

SCR_LINES_pt TXT_GetScrLineByRow(TEXT_BUF_pt atbuf,int r)
{
  int h = atbuf->awp->awin->data_height;
  SCR_LINES_pt aline;

  if (r >= 0) {
    aline = atbuf->scr_lines;
    while (aline && h-- > 0) {
      if (aline->row == r) break;
      if (h > 0) aline++;
      else aline = (SCR_LINES_pt)NULL;
    }
  } else
    aline = (SCR_LINES_pt)NULL;

  return(aline);
}

SCR_LINES_pt TXT_GetScrLine(TEXT_BUF_pt atbuf,long pos)
{
  SCR_LINES_pt aline;
  int h = atbuf->awp->awin->data_height;

  if (pos > atbuf->eop_pos || pos < atbuf->bop_pos)
    aline = (SCR_LINES_pt)NULL;
  else
    aline = atbuf->scr_lines;

  while (aline && h-- > 0) {
    if (pos >= aline->char_pos) {
      if (pos >= atbuf->fsize && pos == aline->char_pos)
	break;
      else {
	long eol = aline->char_pos + aline->len;
	if (pos <= eol) {
	  if (pos < eol ||
	      (!TF_ISSET(aline->flags, EDT_F_CRLF) && eol >= atbuf->fsize))
	      break;
	}
      }
    }
    if (h <= 0) aline = (SCR_LINES_pt)NULL;
    else aline++;
  }

  if (aline && ((aline->char_pos + aline->len) < pos))
    aline = (SCR_LINES_pt)NULL;

  return(aline);
}

void TXT_Write(TEXT_BUF_pt atbuf,long pos,char *buf,int row, int col, int width)
{
  int len = (buf ? strlen(buf) : 0);

  row += w_data_row;
  col += w_data_col;

  if (pos < atbuf->mark_start) {
    int w = imin(width, atbuf->mark_start - pos);
    WIN_Write(atbuf->awp, buf, row, col, w, NORMAL);
    width -= w;
    col += w;
    w = imin(len, w);
    buf += w;
    pos += w;
    len -= w;
  }
  if (width > 0 && pos >= atbuf->mark_start && pos <= atbuf->mark_end) {
    int w = imin(width, atbuf->mark_end - pos);
    if (len <= w && pos < atbuf->fsize) w = width;
    WIN_Write(atbuf->awp, buf, row, col, w, WHTBLU);
    width -= w;
    col += w;
    buf += imin(len, w);
  }

  if (width > 0) WIN_Write(atbuf->awp, buf, row, col, width, NORMAL);
}

void TXT_Redisplay(TEXT_BUF_pt atbuf,SCR_LINES_pt cur_line)
{
  WINFO *w = atbuf->awp->awin;
  int h;
  char buf[STR255LEN];

  if (cur_line == (SCR_LINES_pt)NULL) cur_line = atbuf->scr_lines;

  h = w->data_height - cur_line->row;

  while(h-- > 0) {
    int bytes = EDT_GetBytes(atbuf, (long)cur_line->char_pos,
			     &buf[0], cur_line->len);
    if (bytes > 0 && (buf[bytes - 1] == '\n')) bytes--;
    buf[bytes] = 0;

    if (TF_ISSET(atbuf->flags, TF_DOCUMENT)) {
      if (bytes == 0 && cur_line->len == 1) {
	if (atbuf->hdr_end > 0 && atbuf->hdr_end == cur_line->char_pos) {
	  memset(buf, (int)'.', w->data_width);
	  WIN_Write(atbuf->awp, buf, cur_line->row + w_data_row,
		    w_data_col, w->data_width, NORMAL);
	  cur_line++;
	  continue;
	}
      }
    }

    TXT_Write(atbuf, (long)cur_line->char_pos,
	      buf, cur_line->row, 0, w->data_width);
    cur_line++;
  }
  TXT_DisplayPosition(atbuf);
}

void TXT_UpdateLine(TEXT_BUF_pt atbuf,long pos,int bytes, int size_changed)
{
  SCR_LINES_pt aline = TXT_GetScrLine(atbuf, (long)pos);

  if (size_changed) {
    int old_len;
    int old_row;
    if (!aline) aline = TXT_GetScrLineByRow(atbuf, (int)atbuf->y_cursor);
    if (aline) {
      old_len = aline->len;
      old_row = aline->row;
    } else {
      old_row = 0;
      old_len = 0;
    }
    TXT_RecalcScrLines(atbuf, (long)atbuf->bop_pos);
    aline = TXT_GetScrLine(atbuf, (long)pos);
    if (aline && (aline->len != (old_len + bytes) || aline->row < old_row)) {
      aline = TXT_GetScrLineByRow(atbuf, (int)(imin(old_row, aline->row)));
      TXT_Redisplay(atbuf, aline);
      aline = (SCR_LINES_pt)NULL;
    }
  }

  if (aline) {
    while (bytes > 0) {
      char buf[STR255LEN];
      int offset = pos - aline->char_pos;
      int len = aline->len - offset;
      len = EDT_GetBytes(atbuf, (long)pos, &buf[0], len);
      if (buf[len - 1] == '\n') buf[len - 1] = 0;
      else buf[len] = 0;
      TXT_Write(atbuf, (long)pos, buf, aline->row,
		(int)(pos - aline->char_pos),
		(int)(atbuf->awp->awin->data_width - offset ));
      bytes -= len;
    }
  }
}

void TXT_PositionCursor(TEXT_BUF_pt atbuf,long pos)
{
  SCR_LINES_pt aline = TXT_GetScrLine(atbuf, (long)pos);

  if (aline) {
    int x, y;
    WINFO *w = atbuf->awp->awin;
    atbuf->y_cursor = aline->row;
    atbuf->x_cursor = imin(w->data_width, (pos - aline->char_pos));

    y = atbuf->y_cursor + w->loc.y + w_data_row;
    x = atbuf->x_cursor + w->loc.x + w_data_col;

    if (WIN_FrontWindow(atbuf->awp))
      SCR_SetCursorPosn(y, x);
    else
      WIN_HomeCursor();
  }
}

static int rows_to_move(int rows)
{
  extern int g_LastCmdIndex;

  if (EDT_KeyIndex(KEYCODE_PAGE_DOWN) == g_LastCmdIndex)
    rows--;
  else
    rows = rows / 2 + 1;

  return(rows);
}

long TXT_EstimateStartPos(TEXT_BUF_pt atbuf)
{
  long start_pos = atbuf->char_pos; /* assume where the cursor is */
  SCR_LINES_pt aline;
  int rows = atbuf->awp->awin->data_height;

  if (start_pos >= atbuf->eop_pos) {
    if (rows > 1) {
      aline = TXT_GetScrLineByRow(atbuf, rows_to_move(rows));
      if (aline) start_pos = aline->char_pos;
      else {
	start_pos = imax(0, start_pos - atbuf->awp->awin->data_width);
	aline = TXT_GetScrLine(atbuf, (long)start_pos);
	if (aline) start_pos = aline->char_pos;
      }
    } else
      start_pos = atbuf->char_pos;
  } else if (start_pos < atbuf->bop_pos) {
    long cr_pos = EDT_find_last_eol(atbuf, (long)start_pos);
    int  bytes  = start_pos - cr_pos;
    start_pos = imax(0, imax(start_pos - bytes,
			     start_pos - (atbuf->awp->awin->data_width-1)));
  }
  return(start_pos);
}

void TXT_SetCursor(TEXT_BUF_pt atbuf)
{
  SCR_LINES_pt aline = TXT_GetScrLine(atbuf, (long)atbuf->char_pos);

  if (atbuf->char_pos < atbuf->bop_pos ||
      atbuf->char_pos >= atbuf->eop_pos) {
    if (aline == (SCR_LINES_pt)NULL ||
	(atbuf->char_pos == atbuf->eop_pos &&
	 aline->len >= atbuf->awp->awin->data_width &&
	 !TF_ISSET(aline->flags, EDT_F_CRLF) &&
	 aline->row >= (atbuf->awp->awin->data_height - 1))) {
      long start_pos = atbuf->char_pos;
      if (start_pos == (atbuf->bop_pos - 1) ||
	  start_pos <= (atbuf->eop_pos + atbuf->awp->awin->data_width))
	start_pos = TXT_EstimateStartPos(atbuf);
      TXT_RecalcScrLines(atbuf, (long)start_pos);
      TXT_Redisplay(atbuf, atbuf->scr_lines);
      aline = (SCR_LINES_pt)NULL;
    }
  }

  if (atbuf->mark_start >= 0 && atbuf->mark_end > atbuf->mark_start) {
    if (aline) {
      aline = TXT_GetScrLine(atbuf, (long)atbuf->lastmark);
      TXT_Redisplay(atbuf, aline);
    }
  }

  TXT_PositionCursor(atbuf, (long)atbuf->char_pos);
  SCR_UpdateScreen();
}

TEXT_VIEW *TXT_GetTLByWin(WINLIST *awin)
{
  TEXT_VIEW *atl;

  for (atl=text_list; atl!=NULL; atl=atl->next)
    if (atl->a_tbuf->awp == awin) return(atl);

  return(NULL);
}

TEXT_VIEW *TXT_GetTLForFrontWindow(void)
{
  TEXT_VIEW *atl;

  for (atl=text_list; atl!=NULL; atl=atl->next)
    if (WIN_FrontWindow(atl->a_tbuf->awp)) return(atl);

  return(NULL);
}

void TXT_SetupMenus(int net_state)
{
  TEXT_VIEW *tview = TXT_GetTLForFrontWindow();

  MNU_EnableBar(M_SERVICES, TRUE);

  MNU_EnableItem(M_MAIL_HEADERS, FALSE);

  if (tview) {
    if (TF_ISSET(tview->a_tbuf->flags, TF_CLOSE))
      MNU_EnableItem(M_CLOSE, TRUE);

    if (TF_ISSET(tview->a_tbuf->flags, TF_DOCUMENT)) {
      MNU_EnableItem(M_SEND, TRUE);
      MNU_EnableItem(M_VERIFY, TRUE);
      MNU_EnableItem(M_SAVE_REMOTE, !net_state);
      MNU_EnableItem(M_SAVE, TRUE);
      MNU_EnableItem(M_APND_LOCAL, TRUE);
      MNU_EnableItem(M_APND_REMOTE, !net_state);
      if (TF_ISSET(tview->a_tbuf->flags, TF_EDITING)) {
	MNU_EnableItem(M_MAIL_HEADERS, TRUE);
	MNU_EnableItem(M_ADD_FAXHDRS, (gCanSendFax & (!net_state)));
	MNU_EnableItem(M_ADD_HEADERS, (tview->a_tbuf->hdr_end <= 0));
	MNU_EnableItem(M_CUT, TRUE);
	MNU_EnableItem(M_PASTE, EDT_GetCopyTextLen());
	MNU_EnableItem(M_CLOSE, FALSE);
	MNU_EnableItem(M_APPEND, FALSE);
	MNU_EnableItem(M_SEND, TRUE);
	MNU_EnableItem(M_VERIFY, TRUE);
      } else {
	MNU_EnableItem(M_APPEND, TRUE);
	MNU_EnableItem(M_EDITMSG, TRUE);
	if (gCapturedShell)
	  MNU_EnableItem(M_EXTERNAL_EDIT, FALSE);
	else {
	    const char *tmpdir = config_getstring("TempDir",NULL);
	    MNU_EnableItem(M_EXTERNAL_EDIT, tmpdir != NULL);
	}
      }
    }

    if (tview->a_tbuf->mark_start >= 0 &&
	tview->a_tbuf->mark_end >= tview->a_tbuf->mark_start)
      MNU_EnableItem(M_COPY, TRUE);
    
    MNU_EnableItem(M_SHOW_CONSOLE, TRUE);
    MNU_CheckItem(M_SHOW_CONSOLE, (tview == cons_textview ? TRUE : FALSE));
    
    MNU_EnableItem(M_SHOW_DOCUMENT,
		   (cons_textview->next || cons_textview->prev ?
		    TRUE : FALSE));
    MNU_CheckItem(M_SHOW_DOCUMENT, (tview != cons_textview ? TRUE : FALSE));
  }
}

void TXT_BringNextDocToFront(void)
{
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();

  if (atl) atl = atl->next;
  if (atl == NULL || atl == cons_textview) atl = text_list;
  if (atl) WIN_BringToFront(atl->a_tbuf->awp);
  clear_checks();
  MNU_CheckItem(M_SHOW_DOCUMENT, TRUE);
}

int TXT_Close(WINLIST *w)
{
  TEXT_VIEW *atl = TXT_GetTLByWin(w);
  int flags = atl->a_tbuf->flags;

  if (TF_ISSET(flags, TF_LOCKED)) {
    SCR_Beep();
    return(FALSE);
  }

  if (TF_ISSET(flags, TF_CLOSE)) {
    if (TF_ISSET(flags, TF_FREE)) {
      if (TF_ISSET(flags, TF_DOCUMENT) && TF_ISSET(flags, TF_CHANGED)) {
	if (DLG_GetBooleanFromUser("This document has changed. Close anyway?",
				   FALSE) == FALSE) return(FALSE);
      }
      WIN_Dispose(w);
      TXT_deLink(atl, TRUE);
    } else
      WIN_HideWindow(w);
  }
  return(TRUE);
}

void TXT_RecalcScrLines(TEXT_BUF_pt atbuf,long start_pos)
{
  WINFO *w = atbuf->awp->awin;
  int h = w->data_height;
  SCR_LINES_pt cur_line = atbuf->scr_lines;
  char buf[STR255LEN + 1];

  if (TF_ISSET(atbuf->flags, TF_OFFDSK) && atbuf->fsize > atbuf->bsize) {
    int nbytes = imax(0, start_pos - atbuf->offset);
    if ((atbuf->bsize-nbytes) < (w->data_width * w->data_height)) {
      nbytes = TXT_PageFile(atbuf, (long)(atbuf->offset + nbytes));
    }
  }

  atbuf->bop_pos = start_pos;

  while(h > 0) {
    char *bufptr;
    int bytes = EDT_GetBytes(atbuf, (long)start_pos, &buf[0], STR255LEN);
    if (bytes <= 0) break;
    buf[bytes] = 0;
    bufptr = &buf[0];
    while (bytes > 0 && h > 0) {
      int len;
      cur_line->char_pos = start_pos;
      /* JCS - email shouldn't be wider than this.  The smtp_send() function reformats
       * to this specification, however the editor would otherwise default to 78 columns.
       * This breaks everything else that would like a wider window, but tough luck.
       */
      len = TXT_FormatText ( bufptr, bytes, imin ( atbuf->awp->awin->data_width, 74 ) );
      if (bufptr[len-1] != '\n' && bytes < atbuf->awp->awin->data_width) {
	if ((start_pos + bytes) < atbuf->fsize) {
	  start_pos = cur_line->char_pos;
	  break;
	}
      }
      start_pos += len;
      cur_line->len = len;
      cur_line->row = w->data_height - h;
      cur_line->flags = 0;
      if (bufptr[len-1] == '\n') TF_SET(cur_line->flags, EDT_F_CRLF);
      bytes -= len;
      bufptr += len;
      cur_line++;
      h--;
    }
  }

  atbuf->eop_pos = start_pos;

  while(h > 0) {
    cur_line->char_pos = start_pos;
    cur_line->row = w->data_height - h;
    cur_line->len = 0;
    cur_line->flags = 0;
    cur_line++;
    h--;
  }
}

void TXT_Draw(WINLIST *w)
{
  TEXT_VIEW_pt atl = TXT_GetTLByWin(w);
  TEXT_BUF_pt atbuf = atl->a_tbuf;

  if (TF_ISSET(atbuf->flags, TF_FLOAD)) {
    TXT_LoadFile(atbuf);
  }

  TXT_RecalcScrLines(atbuf, (long)atbuf->bop_pos);
  if (TF_ISSET(atl->a_tbuf->flags, TF_DOCUMENT))
    EDT_RecalcHeaders(atl->a_tbuf);

  WIN_LockCursor((WIN_FrontWindow(w) &&
		  TF_ISSET(atbuf->flags, TF_EDITING)) ? TRUE : FALSE);

  TXT_SyncWin(atbuf);
}

void TXT_Hints(TEXT_BUF_pt atbuf)
{
  if (gNovice && WIN_FrontWindow(atbuf->awp)) {
    int *key_table, key_table_size;
    char *hint_msg;
    if (atbuf->awp == cons_textview->a_tbuf->awp) {
      key_table = Gen_HintKeys;
      key_table_size = sizeof(Gen_HintKeys) / sizeof(int);
      hint_msg = MSG_HintMsgGen;
    } else {
      if (TF_ISSET(atbuf->flags, TF_DOCUMENT)) {
	if (TF_ISSET(atbuf->flags, TF_EDITING)) {
	  key_table = Edt_HintKeys;
	  key_table_size = sizeof(Edt_HintKeys) / sizeof(int);
	  hint_msg = MSG_HintEdtWin;
	} else {
	  if (TF_ISSET(atbuf->flags, TF_AFSFILE)) {
	    key_table = Afs_HintKeys;
	    key_table_size = sizeof(Afs_HintKeys) / sizeof(int);
	    hint_msg = MSG_HintAfsEdit;
	  } else {
	    key_table = Snd_HintKeys;
	    key_table_size = sizeof(Snd_HintKeys) / sizeof(int);
	    hint_msg = (is_net_dead() ? MSG_HintSndNoConnWin : MSG_HintSndWin);
	  }
	}
      } else {
	  key_table = Oth_HintKeys;
	  key_table_size = sizeof(Oth_HintKeys) / sizeof(int);
	  hint_msg = MSG_HintOther;
      }
    }
    HintUser(hint_msg, key_table, key_table_size);
  }
}

void TXT_DoHighlight(WINLIST *win)
{
#ifdef DO_HIGHLIGHT_SELECTION
  TEXT_VIEW *atl = TXT_GetTLByWin(win);
  TXT_SetBanner(atl->a_tbuf, NIL);
#endif /* DO_HIGHLIGHT_SELECTION */
}

void TXT_SetBanner(TEXT_BUF *atbuf,char *aBan)
{
  int attr = BAN_H_ON;
  int col  = 0;
  WINFO *w = atbuf->awp->awin;
  int width = w->data_width;

  TXT_Hints(atbuf);

  if (aBan != NIL) {
    strncpy(atbuf->ban_text, aBan, MAX_BAN);
    if (TF_ISSET(atbuf->flags, TF_CRBLOW))
      strncat(atbuf->ban_text, MSG_ReturnCloses,
	      imax(0, (MAX_BAN - strlen(atbuf->ban_text))));
  }

#ifdef DO_HIGHLIGHT_SELECTION
  if (!WIN_FrontWindow(atbuf->awp)) attr = BAN_H_OFF;
#endif

  if (TF_ISSET(atbuf->flags, TF_CENTER)) {
    int ban_len = strlen(atbuf->ban_text);
    col = imax(0, (width - ban_len)/2);
    WIN_Write(atbuf->awp, NIL, w_ban_row, w_ban_col, width, attr);
    width -= col;
  }
  WIN_Write(atbuf->awp, atbuf->ban_text, w_ban_row,
	    w_ban_col + col, width, attr);

  TXT_DisplayPosition(atbuf);

  WIN_HomeCursor();
}

void TXT_DisplayPosition(TEXT_BUF *atbuf)
{
  WIN_DisplayPosition(atbuf->awp, (long)atbuf->char_pos,
		      (long)atbuf->fsize, FALSE);
}

TEXT_VIEW *TXT_initialize(r, c, h, w, fsize, flags)
int r, c, h, w; 
long fsize;
int flags;
{
  WINLIST  *awp;
  TEXT_BUF *atbuf = (TEXT_BUF *)NULL;
  TEXT_VIEW *atl = (TEXT_VIEW *)NULL;
  const char *tmpdir = config_getstring("TempDir",NULL);

  if (fsize >= PAGEABLE_BUFSIZE && (tmpdir==NULL)) {
    DLG_Alert(1, MSG_FileTooBigToView);
    return(atl);
  }

  awp = WIN_Create(r, c, h, w, TXT_DoHighlight,
		   TXT_DoKey, TXT_Draw, TXT_Close, WTYPE_TEXT);

  if (awp) {
    atl = (TEXT_VIEW *)malloc(sizeof(TEXT_VIEW));
    if (atl) {
      bzero(atl, sizeof(TEXT_VIEW));
      atl->a_tbuf = atbuf = (TEXT_BUF *)malloc(sizeof(TEXT_BUF));
      if (atbuf) {
	unsigned int bufsize;
	bzero(atbuf, sizeof(TEXT_BUF));
	if (TF_ISSET(flags, TF_DOCUMENT)) {
#ifdef FIXED_ALLOCATION
	  if (TF_ISSET(flags, TF_EXACT))
	    bufsize = imin(PAGEABLE_BUFSIZE, fsize);
	  else 
	    bufsize = PAGEABLE_BUFSIZE;
#else
	  bufsize = imin(PAGEABLE_BUFSIZE, fsize);
#endif /* FIXED_ALLOCATION */
	} else
	  bufsize = (unsigned int)fsize;

	atbuf->scr_lines = (SCR_LINES_pt)malloc(
				  (int)(sizeof(SCR_LINES) * screen_rows));
	if (atbuf->scr_lines) {
	  bzero(atbuf->scr_lines, (int)(screen_rows * sizeof(SCR_LINES)));
	  atbuf->buf = (char *)get_deallocable_mem((long)(bufsize + 1));
	}
	if (atbuf->buf) {
	  atbuf->bsize = bufsize;
	  atbuf->ndata = -1L;
	  atbuf->awp = awp;
	  atbuf->flags = flags;
	  atbuf->mark_start = atbuf->mark_end = -1L;
	  TXT_Link(atl);
	}
      }
    }
  }
  if (!awp || !atl || !atbuf || !atbuf->buf) {
    if (atl) {
      if (atbuf) {
	if (atbuf->scr_lines) free(atbuf->scr_lines);
	if (atbuf->buf) free_deallocable_mem(atbuf->buf);
      }
      free(atbuf);
      free(atl);
      atl = (TEXT_VIEW *)NULL;
    }
    if (awp) WIN_dlink(awp, TRUE);
  }

  return(atl);
}

void TXT_deLink(TEXT_VIEW *atl,int dispose)
{
  if (atl->prev) atl->prev->next = atl->next;
  if (atl->next) atl->next->prev = atl->prev;

  if (text_list == atl) text_list = atl->next;

  if (dispose) {
    int flags = atl->a_tbuf->flags;
    if (TF_ISSET(flags, TF_OFFDSK)) {
      if (TF_ISSET(flags, TF_ISOPEN) && atl->a_tbuf->stream)
	FIO_CloseStream(atl->a_tbuf->stream);
      if (TF_ISSET(flags, TF_REMOVE)) FIO_EraseFile(atl->a_tbuf->lname);
    }
#ifndef IBMPC
    else {
      if (atl->a_tbuf->ckpt_file[0]) FIO_EraseFile(atl->a_tbuf->ckpt_file);
    }
#endif
    if (atl->a_tbuf->buf) free_deallocable_mem(atl->a_tbuf->buf);
    if (atl->a_tbuf->scr_lines) free(atl->a_tbuf->scr_lines);
    free(atl->a_tbuf);
    free(atl);
  }
}

void TXT_Link(TEXT_VIEW *atl)
{
  atl->prev = (TEXT_VIEW *)NULL;
  atl->next = text_list;
  if (text_list) text_list->prev = atl;
  text_list = atl;
}

void TXT_trim(TEXT_BUF *aTBuf,int size)
{
  /* trimable text views must fully reside in memory */

  if (size <= 0 || size > aTBuf->bsize) return;

  if (size < aTBuf->bsize) {
    unsigned int nbytes = imax(0, aTBuf->bsize-size);
    strncpy(aTBuf->buf, aTBuf->buf+size, nbytes);
  }
  aTBuf->fsize = imax(0, aTBuf->fsize-size);
}

int TXT_insert(TEXT_BUF *aTBuf,char *aText,char *where,int       text_len)
{
  if (where == aTBuf->buf+aTBuf->fsize) { /* appned ? */
    int nbytes = 0;
    while (text_len-- > 0) {
      *where++ = *aText++;
      nbytes++;
    }
    if (aTBuf->ndata < 0) aTBuf->ndata = aTBuf->fsize;
    if (!TF_ISSET(aTBuf->flags, TF_OFFDSK)) aTBuf->fsize += nbytes;
  }
  return(0);
}

int TXT_append(TEXT_BUF *aTBuf,char     *aText,int       size)
{
  if (!aText || !*aText || size <= 0) return(0);

  if (size >= aTBuf->bsize) {
    if (!TF_ISSET(aTBuf->flags, TF_OFFDSK)) return(-1);
  }

  if (aTBuf->fsize+size >= aTBuf->bsize) {
    if (size >= aTBuf->bsize && !TF_ISSET(aTBuf->flags, TF_OFFDSK))
      return(-1);
    if (TF_ISSET(aTBuf->flags, TF_TRIM)) TXT_trim(aTBuf, size);
    else return(-1);
  }
  if (TXT_insert(aTBuf, aText, aTBuf->buf+aTBuf->fsize, size) != 0) return(-1);
  TXT_SyncWin(aTBuf);
  return(0);
}

int TXT_FormatText(char *text,int len, int width)
{
  int ll, wlen = imin(len, width);

  for (ll = 0; ll < wlen; ll++) {
    if (text[ll] == '\n') return(ll + 1);
  }

  if (wlen == width) {
    while(--wlen > 0) {
      char ch = text[wlen];
      if (ch == SPC_KEY || ch == TAB_KEY) break;
    }
    if (wlen <= 0) wlen = imin(len, width);
    else wlen++;
  }
  return(wlen);
}

void TXT_SyncWin(TEXT_BUF_pt atbuf)
{
  SCR_LINES_pt aline;
  long start_pos = atbuf->bop_pos;
  Boolean scroll_it = FALSE; 

  if (atbuf->ndata >= 0) {
    TXT_RecalcScrLines(atbuf, (long)start_pos);
    if (atbuf->mark_start < 0 && TF_ISSET(atbuf->flags, TF_SCROLL))
      scroll_it = TRUE;
    if (scroll_it) {
      if (atbuf->ndata > atbuf->eop_pos) start_pos = atbuf->ndata;
      atbuf->char_pos = atbuf->ndata;
    }
    atbuf->ndata = -1L;
  }

  if (start_pos != atbuf->bop_pos) {
    TXT_RecalcScrLines(atbuf, (long)start_pos);
    aline = atbuf->scr_lines;
  } else {
    aline = TXT_GetScrLineByRow(atbuf, 0); /* this needs work */
  }

  TXT_Redisplay(atbuf, aline);
  TXT_SetCursor(atbuf);
  if (scroll_it) EDT_EndCursor(atbuf);
  WIN_Update(atbuf->awp);
}

void TXT_MakeBlurb(char *title,int  text_length,char *text_pt)
{
  TEXT_VIEW *atextview;

  atextview = TXT_initialize(5, 2, 15, 76, 0L,
   	             (TF_CLOSE | TF_FREE | TF_CENTER | TF_CRBLOW | TF_EXACT));

  if (!atextview) {
    if (text_pt) free_deallocable_mem(text_pt);
    DLG_Alert(1, MSG_NotEnoughMemory);
  } else {
    TEXT_BUF *tbuf = atextview->a_tbuf;
    TXT_SetBanner(tbuf, title);
    if (tbuf->buf) free_deallocable_mem(tbuf->buf);
    tbuf->buf = text_pt;
    tbuf->bsize = text_length + 1;
    tbuf->fsize = text_length;
    tbuf->buf[tbuf->fsize] = 0;
    WIN_Show(tbuf->awp);
    WIN_BringToFront(tbuf->awp);
  }
}

void TXT_OpenBlurb(char *filename, char *title)
{
  long fsize = FIO_GetFileSize(filename);
  char *text_pt = NULL;

  if (fsize > 0) {
    int rc;
    fsize = imin(fsize, PAGEABLE_BUFSIZE); /* read as much as we can */
    text_pt = (char *)get_deallocable_mem((long)(fsize + 1));
    if (!text_pt) {
      DLG_Alert(1, MSG_NotEnoughMemory);
      return;
    }
    rc = FIO_OpenFile(filename);
    if (rc == 0)
      fsize = FIO_ReadFromFile(text_pt, (int)fsize);
    else
      fsize = -1L;
    if (rc == 0) FIO_CloseFile();
  }

  if (fsize < 0) {
    CON_Write(MSG_OpenFailed);
    CON_WriteLine(filename);
  } else
    TXT_MakeBlurb(title, (int)fsize, text_pt);
}

void TXT_Edit(void)
{
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();

  EDT_Edit(atl->a_tbuf);
}

void TXT_ExtEdit(void)
{
  int rc, changed = FALSE;
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();
  TEXT_BUF  *atbuf = atl->a_tbuf;
  const char *editor = config_getstring("ExternalEditor",NULL);

  if (!(TF_ISSET(atbuf->flags, TF_DOCUMENT))) {
    CON_WriteLine("Front window is not an editable document!");
    return;
  }

  if (!editor) {
    CON_WriteLine("No Editor is set, Set an Editor in local preferences.");
    return;
  }

  if (TF_ISSET(atbuf->flags, TF_ISOPEN)) {
    TF_CLEAR(atbuf->flags, TF_ISOPEN);
    FIO_CloseStream(atbuf->stream);
    atbuf->stream = (FILE *)NULL;
  }

  if (!TF_ISSET(atbuf->flags, TF_OFFDSK)) {
    if ((rc = FIO_GenTmpName(atbuf->lname)) == 0)
      if ((rc = TXT_WriteFile(atbuf, atbuf->lname)) == 0) {
	TF_SET_OFFDSK(atbuf->flags);
	TF_SET_REMOVE(atbuf->flags);
      }
    if (rc != 0) {
      DLG_Alert(1, MSG_WriteError);
      return;
    }
  }
  if (atbuf->buf) {
    free_deallocable_mem(atbuf->buf);
    atbuf->buf = NIL;
  }

  atbuf->fsize = 0;
  atbuf->bsize = 0;
  atbuf->offset = 0;
  atbuf->char_pos = 0;
  atbuf->mark_start = -1L;
  atbuf->mark_end = -1L;
  atbuf->eop_pos = 0;
  atbuf->bop_pos = 0;

  rc = FIO_EditFile(atbuf->lname, &changed);

  if (rc) {
    CON_Write("Error editing the document with editor ");
    CON_WriteLine((char *) editor);
  }
  else
    if (changed) TF_SET_CHANGED(atbuf->flags);

  atbuf->fsize = FIO_GetFileSize(atbuf->lname);
  TXT_ReallocTview(atl, (long)atbuf->fsize);

  if (atbuf->fsize <= atbuf->bsize) {
    if (DOC_ReadFile(atbuf, atbuf->lname, (long)atbuf->fsize) < 0)
      DLG_Alert(1, MSG_ReadError);
    else {
      TF_CLEAR(atbuf->flags, TF_OFFDSK);
      if (TF_ISSET(atbuf->flags, TF_REMOVE)) FIO_EraseFile(atbuf->lname);
    }
  }

  if (TF_ISSET(atbuf->flags, TF_OFFDSK)) TF_SET_FLOAD(atbuf->flags);

  TXT_Draw(atbuf->awp);
}

int TXT_WriteFile(TEXT_BUF *atbuf,char *fname)
{
  int rc;
  long len = atbuf->fsize;

  if (TF_ISSET(atbuf->flags,TF_DOCUMENT) && TF_ISSET(atbuf->flags,TF_OFFDSK)) {
    if (TF_ISSET(atbuf->flags, TF_ISOPEN)) {
      FIO_CloseStream(atbuf->stream);
      atbuf->stream = (FILE *)NULL;
      TF_CLEAR(atbuf->flags, TF_ISOPEN);
    }
    if ((rc = FIO_CloneFile(atbuf->lname, fname, (long)len)) == 0) {
      DOC_LoadLocalFile(fname, (long)atbuf->fsize, NIL, FALSE,FALSE);
      TF_CLEAR(atbuf->flags, TF_CHANGED);
      TXT_Close(atbuf->awp);
    }
  } else {
    char *buf = atbuf->buf;
    if ((rc = FIO_CreateFile(fname)) == 0) {
      while (len > 0 && (rc = FIO_WriteToFile(buf, (unsigned int)len)) > 0) {
	buf += rc;
	len -= rc;
	rc = 0;
      }
      FIO_CloseFile();
    }
  }
  if (rc != 0) {
    CON_Write(MSG_WriteError);
    CON_WriteLine(fname);
  }
  return(rc);
}

int TXT_DoMenuCommand(int m_id,char *str)
{
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();

  if (atl == NULL) return(FALSE);

  if (TF_ISSET(atl->a_tbuf->flags, TF_LOCKED))
    SCR_Beep();
  else {
    switch(m_id) {
    case M_SAVE_REMOTE:
      if (TF_ISSET(atl->a_tbuf->flags, TF_DOCUMENT))
	DOC_SaveAsRemote(atl, str);
      break;
    case M_SAVE:
    case M_SAVE_LOCAL:
      if (TXT_WriteFile(atl->a_tbuf, str) == 0) {
	CON_Write(MSG_FileSavedIn); CON_WriteLine(str);
	TF_CLEAR(atl->a_tbuf->flags, TF_CHANGED);
	WIN_SetTitle(atl->a_tbuf->awp, str);
      }
      break;
    case M_APND_REMOTE:
      DOC_AppendRemote(atl, str);
      break;
    case M_APPEND:
    case M_APND_LOCAL:
      DOC_AppendLocal(atl, str);
      break;
    case M_SEND:
      DOC_VerifyMsg(atl);
      DOC_SubmitMsg(atl);
      break;
    case M_VERIFY:
      DOC_VerifyMsg(atl);
      break;
    case M_SIGNMSG:
	DOC_signmessage(atl);		  
	break;
    default:
      return(FALSE);
    }
  }
  return(TRUE);
}

int TXT_ReallocTview(TEXT_VIEW *atl,long new_size)
{
  char *new_buf;
  TEXT_BUF *tview = atl->a_tbuf;

  if (TF_ISSET(tview->flags, (TF_DOCUMENT | TF_OFFDSK))) {
#ifdef    FIXED_ALLOCATION
    new_size = PAGEABLE_BUFSIZE;
#else  /* FIXED_ALLOCATION */
    new_size = imin(new_size, PAGEABLE_BUFSIZE);
#endif /* FIXED_ALLOCATION */
  }

  if (atl->a_tbuf->buf) {
    if (new_size != tview->bsize)
      new_buf = realloc_deallocable_mem(tview->buf, (long)(new_size + 1),
					(long)tview->bsize);
    else
      new_buf = tview->buf;
  } else
    new_buf = get_deallocable_mem((long)(new_size + 1));

  if (new_buf) {
    tview->buf = new_buf;
    tview->bsize = new_size;
    tview->ndata = -1L;
  }
  return(new_buf ? 0 : -1);
}

int TXT_GetData(TEXT_BUF *tbuf,
		char *buf,
		int len,
		long offset)
{
  if (offset < tbuf->fsize && offset >= 0) {
    if (TF_ISSET(tbuf->flags, TF_OFFDSK)) {
      if (!TF_ISSET(tbuf->flags, TF_ISOPEN)) {
	tbuf->stream = FIO_OpenStream(tbuf->lname, OPEN_R);
	TF_SET_ISOPEN(tbuf->flags);
      }
      if (FIO_SeekStream(tbuf->stream, (long)offset) == 0)
	len = FIO_ReadStream(tbuf->stream, buf, len);
      else
	len = -1;
    } else { /* all in memory */
      len = imax(0, (imin(len, (tbuf->fsize - offset))));
      if (len > 0) bcopy(&tbuf->buf[offset], buf, len);
    }
    return(len);
  }
  return(-1);
}

void TXT_LoadFile(TEXT_BUF *tbuf)
{
  int rc = 0;

  if (!tbuf->stream || !TF_ISSET(tbuf->flags, TF_ISOPEN))
    tbuf->stream = FIO_OpenStream(tbuf->lname, OPEN_R);
  if (tbuf->stream) {
    char *bufptr = tbuf->buf;
    unsigned int len = imin(tbuf->fsize, tbuf->bsize);
    while (len > 0) {
      rc = FIO_ReadStream(tbuf->stream, bufptr, len);
      if (rc <= 0) break;
      bufptr += rc;
      len    -= rc;
      *bufptr = '\0';
    }
  }
  if (rc < 0 || !tbuf->stream) DLG_Alert(1, MSG_ReadError);
  else TF_SET_ISOPEN(tbuf->flags);
  TF_CLEAR(tbuf->flags, TF_FLOAD);
  TXT_DisplayPosition(tbuf);
}

int TXT_PageFile(TEXT_BUF *tbuf,long offset)
{
  /* read data from the files associated with this text buffer.
     read starts from offset up to max buffer size.
     This routine tries to fill the buffer, so offset may not
     correspond to the beggining of the buffer.
  */

  char *fillptr;
  int nbytes, bytes_to_read;

  if (!TF_ISSET(tbuf->flags, (TF_DOCUMENT | TF_OFFDSK))) return(-1);

  if ((offset >= tbuf->fsize) || (tbuf->fsize <= tbuf->bsize) ||
      (offset >= tbuf->offset && tbuf->fsize <= (tbuf->offset + tbuf->bsize)))
    return(0); /* EOF */

  if (!TF_ISSET(tbuf->flags, TF_ISOPEN)) { /* file is not open */
    if (!(tbuf->stream = FIO_OpenStream(tbuf->lname, OPEN_R))) return(-1);
    TF_SET_ISOPEN(tbuf->flags);
  }

  bytes_to_read = imin(tbuf->bsize, iabs(tbuf->offset - offset));

  if (bytes_to_read == 0) return(0);

  if (bytes_to_read < tbuf->bsize) { /* can preserve part of buffer */
    char *from, *to;
    bytes_to_read = imin(bytes_to_read,
			 (bytes_to_read +
			  (tbuf->fsize - (offset + tbuf->bsize))));
    nbytes = tbuf->bsize - bytes_to_read;
    if (offset < tbuf->offset) {
      from = tbuf->buf;
      to   = tbuf->buf + bytes_to_read;
      tbuf->offset -= bytes_to_read;
      offset = imax(0, tbuf->offset - nbytes);
      fillptr = tbuf->buf;
    } else {
      from = tbuf->buf + bytes_to_read;
      to   = tbuf->buf;
      tbuf->offset += bytes_to_read;
      offset = tbuf->offset + nbytes;
      fillptr = tbuf->buf + nbytes;
    }
    memmove(to, from, nbytes);
  } else {
    fillptr = tbuf->buf;
    tbuf->offset = imin(offset, (tbuf->fsize - tbuf->bsize));
    offset = tbuf->offset;
  }

  if ((nbytes = FIO_SeekStream(tbuf->stream, (long)offset)) == 0)
    nbytes = FIO_ReadStream(tbuf->stream, fillptr, bytes_to_read);
    
  return(nbytes);
}

int TXT_DoKey(int key)
{
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();
  TEXT_BUF *atbuf = atl->a_tbuf;

  if ((key == KEYCODE_RETURN) || (key == KEYCODE_FAKERETURN)) {
    if (atbuf->mark_start >= 0) {
      atbuf->mark_start = atbuf->mark_end = -1L;
      TXT_Draw(atbuf->awp);
      SCR_WriteStatusLine(MSG_MarkCleared);
      return(TRUE);
    } else if (TF_ISSET(atbuf->flags, TF_CRBLOW))
      return(DoMenuCommand(M_CLOSE));
  }

  return(EDT_ProcessKey(key, atbuf));
}

void TXT_CopyText(void)
{
  extern void EDT_Copy();
  TEXT_VIEW_pt tview = TXT_GetTLForFrontWindow();

  if (tview) EDT_Copy(tview->a_tbuf);
}

int TXT_PrintWindow(void)
{
  TEXT_VIEW *atl = TXT_GetTLForFrontWindow();

  if (atl && PRT_OpenPrinter()) {
    TEXT_BUF *atbuf = atl->a_tbuf;
    long offset;
    int len;
    offset = 0L;
    while(offset < atbuf->fsize) {
      char buf[1024];
      if (SCR_Keyboard_io()) {
	while (SCR_Keyboard_io()) SCR_GetKey();
	if (DLG_GetBooleanFromUser("Stop Printing?", TRUE)) break;
      }
      len = TXT_GetData(atbuf, buf, sizeof(buf)-1, (long)offset);
      if (len <= 0) break;
      PRT_WriteBuf(buf, len);
      offset += len;
    }
    PRT_Close();
  }
  return(TRUE);
}

void TXT_ReportSize(void)
{
  TEXT_VIEW_pt tview = text_list;

  while (tview) {
    if (tview != cons_textview) {
      TEXT_BUF_pt atbuf = tview->a_tbuf;
      if (TF_ISSET(atbuf->flags, TF_DOCUMENT) && WIN_isShown(atbuf->awp)) {
	


	/* 
WINFO *w = atbuf->awp->awin;
xxx a_local_prefs->loc[SEND_RL].y = w->loc.y;
	a_local_prefs->loc[SEND_RL].x = w->loc.x;
	a_local_prefs->dim[SEND_RL].y = w->dim.y;
	a_local_prefs->dim[SEND_RL].x = w->dim.x; */
	break;
      }
    }
    tview = tview->next;
  }
}

